# frozen_string_literal: true

require "core/local"

module Refine
  module Object
    module Copy
      DEFAULT = ::Object.new

      refine ::Object do
        # [public] Copies the object using `clone`.
        #
        if RbConfig::CONFIG["RUBY_PROGRAM_VERSION"] < "3"
          def copy(freeze: DEFAULT)
            should_freeze = resolve_freeze_argument(freeze)

            value = clone(freeze: should_freeze)
            value.freeze if should_freeze
            value
          end
        else
          def copy(freeze: DEFAULT)
            clone(freeze: resolve_freeze_argument(freeze))
          end
        end

        private def resolve_freeze_argument(value)
          case value
          when DEFAULT
            frozen?
          else
            !!value
          end
        end
      end

      refine Array do
        # [public] Copies the array, along with each value.
        #
        def copy(freeze: DEFAULT)
          unless Copy.copying?(self)
            Copy.prevent_recursion(self) do
              array = map { |value|
                value.copy(freeze: freeze)
              }

              array.freeze if resolve_freeze_argument(freeze)

              array
            end
          end
        end
      end

      refine Hash do
        # [public] Copies the hash, along with each key and value.
        #
        def copy(freeze: DEFAULT)
          unless Copy.copying?(self)
            Copy.prevent_recursion(self) do
              hash = {}

              each_pair do |key, value|
                hash[key.copy(freeze: freeze)] = value.copy(freeze: freeze)
              end

              hash.freeze if resolve_freeze_argument(freeze)

              hash
            end
          end
        end
      end

      class << self
        include Is::Localized

        def prevent_recursion(object)
          object_id = object.object_id
          copied_objects[object_id] = true
          yield
        ensure
          copied_objects.delete(object_id)
        end

        def copying?(object)
          copied_objects[object.object_id]
        end

        def copied_objects
          localized(:__corerb_copied_objects) || localize(:__corerb_copied_objects, {})
        end
      end
    end
  end
end
